/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice
 * (including the next paragraph) shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL VIA, S3 GRAPHICS, AND/OR
 * ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "via_chrome9_hdac_patch.h"


#define HDAC_PATCH_TIME_MS    20

/* HDMI CTS patch variable */
typedef struct _HdmiPatch {
	u32 CTS;
	u32 Interval;
	u32 Feedback;
	u32 Counter;
	u32 Increaser;

	u32 PixelClockSaved;
	u32 HdacSampleRateSaved;
	u8 isIga2Hdmi;

	struct timer_list ctsPatchTimerVar;

} HdmiPatchRec, *HdmiPatchPtr;

static HdmiPatchRec hdacPatchVar = {0, 0, 0, 0, 0, 0, 0, 0};

static u8 hdac_timer_open = {0};
static u32 debug_counter = {0};



/*
 * The function implement 64bit division. Because Linux do not support 64bit
 * division in linxu kernal space.
 * d:divisor;  n:dividend; N:bits of quotient.
 * example : a(40bit) = b / c;    =>   a = via_chrome9_udiv(c, b, 40);
 */
static unsigned long long via_chrome9_udiv(unsigned long long d,
		unsigned long long n, unsigned long N) {
	unsigned long long q = 0, r = n;
	/* calculate next quotient bit */
	do {
		/* move to next bit */
		N--;
		/* if r >= d * (1 << N) */
		if ((r >> N) >= d) {
			/* update remainder */
			r -= (d << N);
			/* update quotient */
			q += (1LL << N);
		}
	} while (N);
	return q;
}

static u16 get_chipset_version(void)
{
	u16 ret = 0;
	ret = via_chrome9_read_vga_io(REG_SR37);
	ret |= (u16)(via_chrome9_read_vga_io(REG_SR38)) << 8;
	return ret;
}

static u8 internal_hdmi_Iga_path(void)
{
	#define HDTV0_ON_IGA2   0x02

	u8 regCRFF = via_chrome9_read_vga_io(REG_CRFF);

	if (HDTV0_ON_IGA2 & regCRFF)
		return IGA2;
	else
		return IGA1;

}

static u8 is_iga_downscale_enable(int igaPath)
{
	u8 regCR89;

	if (IGA1 == igaPath) {
		via_chrome9_write_vga_io_bits(REG_CRFD, BIT7, BIT7);
		regCR89 = via_chrome9_read_vga_io(REG_CR89);

		if (BIT0 & regCR89)
			return 1;
		else
			return 0;
	} else {
		via_chrome9_write_vga_io_bits(REG_CRFD, 0x00, BIT7);
		regCR89 = via_chrome9_read_vga_io(REG_CR89);

		if (BIT0 & regCR89)
			return 1;
		else
			return 0;
	}

}

/* This function transfer MRN register bytes to Clock Frequency(Hz) */
static u32 calc_clk_from_reg(u8 regSR47, u8 regSR48, u8 regSR49)
{
	#define UMA_CLK_Reference          14318180
	/**
	* Clcok Synthesizer Value 0 :      DM[9:0]
	* Clcok Synthesizer Value 1[4:2] : DR[2:0]
	* Clcok Synthesizer Value 2[6:0] : DN[6:0]
	*/
	u32 PLLDM, PLLDN, PLLDR;
	u32 ret;

	PLLDM = (u32)(regSR48 & 0x03);
	PLLDM <<= 8 ;
	PLLDM |= (u32)regSR47;
	PLLDR = (u32)((regSR48 & 0x1C) >> 2);
	/* Remove DTZ bit */
	PLLDN = (u32)(regSR49 & 0x7F);

	if (PLLDN == 0)
		return 0;

	/* Notes: we only support for the chipset after VX800 */
	switch (get_chipset_version()) {
	case VX800_DEVICE:
		/* VX800 need to add 2 before calculate them to pixelclock. */
		PLLDM += 2;
		PLLDN += 2;
		break;
	case VX855_DEVICE:
	case VX900_DEVICE:
	default:
		break;
	}

	/* True Output Fout = Fref * (DM) / [(DN)(2 ^ DR)] */
	ret = (UMA_CLK_Reference * (PLLDM) / ((PLLDN) * ((u32)1 << PLLDR)));

	return ret;
}


/* Get Pixel Clock from IGAn,  Frequency(Hz) */
static u32 via_get_pixel_clock(int igaPath)
{
		u8 regSR47, regSR48, regSR49;
		u32 ret = 0;

	if (IGA2 == igaPath) {
		regSR47 = via_chrome9_read_vga_io(REG_SR4A);
		regSR48 = via_chrome9_read_vga_io(REG_SR4B);
		regSR49 = via_chrome9_read_vga_io(REG_SR4C);
	} else {
		regSR47 = via_chrome9_read_vga_io(REG_SR44);
		regSR48 = via_chrome9_read_vga_io(REG_SR45);
		regSR49 = via_chrome9_read_vga_io(REG_SR46);
	}

	ret = calc_clk_from_reg(regSR47, regSR48, regSR49);

	/* if do down scaling, the pixel clock should be divided by 2  */
	if (is_iga_downscale_enable(igaPath))
		ret = ret >> 1;

	return ret;
}


static u32 via_get_sample_rate(void)
{
	u32 mmC4D4, mmC4D8, hdacSampleRate = 0;

	/* step 1. read out C4D8 */
	mmC4D4 = MMIO_RD(0xC4D4) & 0x00FFFFFF;
	/* stream format */
	mmC4D4 |=  0x01000000;
	MMIO_WR(0xC4D4, mmC4D4);

	/* [15:0] */
	mmC4D8 = MMIO_RD(0xC4D8);

	/*
	* step 2. decode C4D8 by definition of C45C
	* C45C [14] Converter_Format_Base
	* BASE. Converter Format Sample Base Rate. (PCM Format structure bit 14)
	*              0: 48kHz
	*              1: 44.1kHz
	* C45C [13:11] Converter_Format_Mult
	* MULT. Converter Format Sample Base Rate Multiple.
	(PCM Format structure bits 13:11)
	*              000: 48kHz/ 44.1 kHz or lesss
	*              001: x2 (96kHz, 88.2kHz, 32kHz)
	*              010: x3 (144kHz)
	*              011: x4 (192kHz, 176.4kHz)
	*              100-111: Reserved
	* C45C [10:8] Converter_Format_Div
	* DIV. Converter Format Sample Base Rate Divisor.
	(PCM Format structure bits 10:8)
	*              000: Divide by 1 (48kHz, 44.1 kHz)
	*              001: Divide by 2 (24kHz, 22.05kHz)
	*              010: Divide by 3 (16kHz, 32kHz)
	*              011: Divide by 4 (11.025kHz)
	*              100: Divide by 5 (9.6kHz)
	*              101: Divide by 6 (8kHz)
	*              110: Divide by 7
	*              111: Divide by 8 (6kHz)
	*/
	if (mmC4D8 & 0x4000)
		hdacSampleRate = 44100;
	else
		hdacSampleRate = 48000;

	switch ((mmC4D8 & 0x3800) >> 11) {
	case 1:
		hdacSampleRate *= 2;
		break;
	case 2:
		hdacSampleRate *= 3;
		break;
	case 3:
		hdacSampleRate *= 4;
		break;
	default:
		break;
	}

	switch ((mmC4D8 & 0x700) >> 8) {
	case 1:
		hdacSampleRate /= 2;
		break;
	case 2:
		hdacSampleRate /= 3;
		break;
	case 3:
		hdacSampleRate /= 4;
		break;
	case 4:
		hdacSampleRate /= 5;
		break;
	case 5:
		hdacSampleRate /= 6;
		break;
	case 6:
		hdacSampleRate /= 7;
		break;
	case 7:
		hdacSampleRate /= 8;
		break;
	default:
		break;
	}

	return hdacSampleRate;
}



static u8 via_reset_cts_patch_variable(
		u32 hdacSampleRate,  u32  hdacPixelClock,
		HdmiPatchPtr  pHdacPatch)
{
	unsigned long long N = 0, CTS = 0, CTSx1000 = 0;

	if (hdacPixelClock == 0 || hdacSampleRate == 0)
		return 0;

	/* Remove the comment when upstream!! */
	/* N = (128 * hdacSampleRate)/1000, fix N value to 6144
	* CTS = PixelClock * N / (128 * hdacSampleRate)
	*/
	N = 6144;
	CTS = (unsigned long long)(hdacPixelClock) * N;
	CTSx1000 = CTS * 1000;
	CTS = via_chrome9_udiv((unsigned long long)(128*hdacSampleRate),
		CTS, 32);
	CTSx1000 = via_chrome9_udiv((unsigned long long)(128*hdacSampleRate),
		CTSx1000, 32);

	pHdacPatch->Interval = (u32)(CTSx1000 - (CTS * 1000));
	pHdacPatch->CTS = (u32)CTS;
	pHdacPatch->Feedback = 0;
	pHdacPatch->Counter = 0;
	pHdacPatch->Increaser = 0;

	return 1;

}


u8 via_chrome9_set_HDAudio_CTS_N(u32 hdacSampleRate,
		u32 hdacPixelClock)
{
	u32  mmC294 = 0, mmC298 = 0, mmC29C = 0;
	unsigned long long int  CTS = 0, N = 0;

	if (hdacPixelClock == 0 || hdacSampleRate == 0)
		return 0;

	/* Remove the comment when upstream!! */
	/* N = (128 * hdacSampleRate)/1000, fix N value to 6144
	* CTS = PixelClock * N / (128 * hdacSampleRate)
	*/
	N = 6144;
	CTS = (unsigned long long)(hdacPixelClock) * N;
	CTS = via_chrome9_udiv((unsigned long long)(128*hdacSampleRate),
		CTS, 32);

	KMS_DEBUG("CTS=%llx\n", CTS);

	mmC294 = (MMIO_RD(0xc294) & 0x0000000F) | 0x100000C0;
	/* notes: C294[27:8] shoule be the same value as cts */
	mmC294 |= ((u32)(CTS) << 8);
	MMIO_WR(0xC294, mmC294);

	mmC298 = (u32)(N) | (((u32)(CTS) & 0xFFF) << 20);
	MMIO_WR(0xC298, mmC298);

	mmC29C = MMIO_RD(0xC29C) & 0xFFFFFF00;
	mmC29C |= (((u32)(CTS) >> 12) & 0xFF);
	MMIO_WR(0xC29C, mmC29C);

	return 1;
}


u8 via_chrome9_set_HDAudio_sample_rate(u32 hdacSampleRate,
		u32 hdacPixelClock)
{
	u32  mmC400 = 0, mmC404 = 0, mmC410 = 0, mmC414 = 0;
	unsigned long long hdacWallClock = 0, hdacPacketClock = 0;
	unsigned long long two2fourty = 1099511627776LL;  /* 2^40 */

	if (hdacPixelClock == 0 || hdacSampleRate == 0)
		return 0;

	mmC404 = MMIO_RD(0xC404) & 0xFFFFFF00;
	mmC414 = MMIO_RD(0xC414) & 0xFFFFFF00;

	/* Remove the comment when upstream!! */
	/* set Wall Clock = (two2fourty / hdacPixelClock)* 24M Hz */
	hdacWallClock = via_chrome9_udiv(hdacPixelClock, two2fourty, 40);
	hdacWallClock *= 24000000;
	mmC400 = (u32)(hdacWallClock & 0xFFFFFFFF);
	mmC404 |= (u32)(hdacWallClock >> 32);

	MMIO_WR(0xC404, mmC404);
	MMIO_WR(0xC400, mmC400);

	/* Remove the comment when upstream!! */
	/* set audio sample rate:
	hdacSampleRate * two2fourty / hdacPixelClock */
	hdacPacketClock = hdacSampleRate * two2fourty;
	hdacPacketClock = via_chrome9_udiv(hdacPixelClock, hdacPacketClock, 40);
	mmC410 = (u32)(hdacPacketClock & 0xFFFFFFFF);
	mmC414 |= (u32)(hdacPacketClock >> 32);

	MMIO_WR(0xC414, mmC414);
	MMIO_WR(0xC410, mmC410);

	KMS_DEBUG("hdacWallClock=%llx, hdacPacketClock=%llx\n",
		hdacWallClock, hdacPacketClock);
	KMS_DEBUG("mmC400=%x, mmC404=%x, mmC410=%x, mmC414=%x\n",
		mmC400, mmC404, mmC410, mmC414);

	return 1;
}


static void via_change_hdmi_audio_env(HdmiPatchPtr pHdacPatch)
{
	u32 mmC424/*, mmC4D4*/;
	u32 hdacSampleRate, hdacPixelClock = 0;

	hdacPixelClock = pHdacPatch->PixelClockSaved;
	hdacSampleRate = pHdacPatch->HdacSampleRateSaved;

	if ((!hdacPixelClock)  || (!hdacSampleRate))
		return;

	mmC424 = MMIO_RD(0xC424) & 0xFFFF9FFF;
	/*mmC424 |= (0 << 13);*/
	MMIO_WR(0xC424, mmC424);

	KMS_DEBUG("hdacSampleRate=%x, hdacPixelClock=%x\n",
		hdacSampleRate, hdacPixelClock);

	/* Remove the comment when upstream!! */
	/* set audio sample rate: register C400 and C410*/
	via_chrome9_set_HDAudio_sample_rate(hdacSampleRate, hdacPixelClock);
	
	/* Remove the comment when upstream!! */
	/* fill CTS & N according to Table7-1, 7-2, 7-3 of HDMI spec 1.3 */
	via_chrome9_set_HDAudio_CTS_N(hdacSampleRate, hdacPixelClock);
	
	/* Remove the comment when upstream!! */
	/* reset the value of cts patch variables. */
	via_reset_cts_patch_variable(hdacSampleRate, hdacPixelClock,
		pHdacPatch);

}

static void via_hdmi_cts_patch(HdmiPatchPtr pHdacPatch)
{
	u32 CTSnow = 0, CTS = 0, mmC294 = 0, mmC298 = 0, mmC29C = 0;
	u32 hdacPixelClock;
	/* patch it 50 times per second. */
	u32 Total = 50;

	/* get refresh rate from read registers. */
	hdacPixelClock = pHdacPatch->PixelClockSaved;

	CTSnow = (MMIO_RD(0xC294) & 0x0FFFFF00) >> 8;
	CTS = CTSnow;

	if (CTSnow != pHdacPatch->CTS
		&& (pHdacPatch->Counter * 1000) / Total >=
		(pHdacPatch->Interval - pHdacPatch->Feedback)) {
		pHdacPatch->Increaser = ((pHdacPatch->Counter * 1000) / Total) -
		pHdacPatch->Interval + pHdacPatch->Feedback;
		CTS = pHdacPatch->CTS;
	}

	if (CTSnow != pHdacPatch->CTS + 1
		&& ((pHdacPatch->Counter * 1000) / Total <
		(pHdacPatch->Interval - pHdacPatch->Feedback))) {
		CTS = pHdacPatch->CTS + 1;
	}

	pHdacPatch->Counter = (pHdacPatch->Counter + 1) % Total;

	if (pHdacPatch->Counter == 0)
		pHdacPatch->Feedback = pHdacPatch->Increaser;

	if (CTS != CTSnow) {
		mmC294 = (MMIO_RD(0xC294) & 0x0000000F) | 0x100000C0;
		mmC294 |= (CTS << 8);
		MMIO_WR(0xC294, mmC294);

		mmC298 = MMIO_RD(0xC298) & 0x000FFFFF;
		mmC298 |= ((CTS & 0xFFF) << 20);
		MMIO_WR(0xC298, mmC298);

		mmC29C |= (CTS >> 12) & 0xFF;
		MMIO_WR(0xC29C, mmC29C);
	}
}

static u8 is_hdmi_audio_env_changed(HdmiPatchPtr pHdacPatch)
{
	u32 hdacPixelClock, hdacSampleRate;
	u8 isIga2HdmiTemp, ret = 0;

	isIga2HdmiTemp = internal_hdmi_Iga_path();
	if (pHdacPatch->isIga2Hdmi != isIga2HdmiTemp) {
		pHdacPatch->isIga2Hdmi = isIga2HdmiTemp;
		ret |= BIT0;
	}

	hdacPixelClock = via_get_pixel_clock(isIga2HdmiTemp);
	hdacSampleRate = via_get_sample_rate();
	if ((!hdacPixelClock) || (!hdacSampleRate))
		return 0;

	if (pHdacPatch->PixelClockSaved != hdacPixelClock) {
		pHdacPatch->PixelClockSaved = hdacPixelClock;
		ret |= BIT1;
	}

	if (pHdacPatch->HdacSampleRateSaved != hdacSampleRate) {
		pHdacPatch->HdacSampleRateSaved = hdacSampleRate;
		ret |= BIT2;
	}

	if (ret != 0) {
		via_change_hdmi_audio_env(pHdacPatch);
		KMS_DEBUG("hdmi audio environment changed! ret =0x%x\n", ret);
	}

	return ret;
}

static void via_hdmi_cts_patch_timer_taskle(void)
{
	#define PHY_FUNCTION_SELECT_HDMI    0x01
	#define HDMI_MODE_ENBLE                    0x00000002
	#define HDMI_MODE_ENBLE_MASK          0x00000042

	/* bak 0x3c4 (SR_REG index) */
	u8 reg_3c4_save = STANDVGA_R8(REG_ZSR);
	/* bak 0x3d4 (CR_REG index) */
	u8 reg_3d4_save = STANDVGA_R8(REG_ZCR);

	u32 HdmiCtlFlag = MMIO_RD(0xC280);
	u8 regCRFF = via_chrome9_read_vga_io(REG_CRFF);

	/* We run cts patch just when internal HDMI is working. */
	if ((PHY_FUNCTION_SELECT_HDMI & regCRFF) &&
		(HDMI_MODE_ENBLE == (HdmiCtlFlag & HDMI_MODE_ENBLE_MASK))) {
		is_hdmi_audio_env_changed(&hdacPatchVar);
		via_hdmi_cts_patch(&hdacPatchVar);
	}

	/* restore the SR_REG index */
	STANDVGA_W8(REG_ZSR, reg_3c4_save);
	/* restore the CR_REG index */
	STANDVGA_W8(REG_ZCR, reg_3d4_save);
}


static void via_chrome9_hdac_cts_patch_timer_func(unsigned long arg)
{
	int *pdata = (int *)arg;
	struct timer_list *ctsPatchTimerPtr = &(hdacPatchVar.ctsPatchTimerVar);

	if ((*pdata)%500 == 0)
		KMS_DEBUG("cts patch runs %d seconds past.\n", (*pdata)/50);

	(*pdata)++;
	via_hdmi_cts_patch_timer_taskle();
	ctsPatchTimerPtr->expires = get_jiffies_64()+msecs_to_jiffies(
		HDAC_PATCH_TIME_MS);
	/* register timer continue. */
	add_timer(ctsPatchTimerPtr);
}

static void via_chrome9_hdac_cts_patch_timer_init(void)
{
	struct timer_list *ctsPatchTimerPtr = &(hdacPatchVar.ctsPatchTimerVar);
	init_timer(ctsPatchTimerPtr);
	ctsPatchTimerPtr->expires = get_jiffies_64()+msecs_to_jiffies(
		HDAC_PATCH_TIME_MS);
	ctsPatchTimerPtr->data = (unsigned long)&debug_counter;
	ctsPatchTimerPtr->function = (void *)
		via_chrome9_hdac_cts_patch_timer_func;
	add_timer(ctsPatchTimerPtr);
	KMS_DEBUG("hdac cts patch run!\n");
}

static void via_chrome9_hdac_cts_patch_timer_exit(void)
{
	struct timer_list *ctsPatchTimerPtr = &(hdacPatchVar.ctsPatchTimerVar);

	del_timer_sync(ctsPatchTimerPtr);

	/* Clear the cts patch variables,
	ready for re-enable cts patch timer. */
	hdacPatchVar.PixelClockSaved = 0;
	hdacPatchVar.HdacSampleRateSaved = 0;
	hdacPatchVar.isIga2Hdmi = 0;
	hdacPatchVar.CTS = 0;
	hdacPatchVar.Interval = 0;
	hdacPatchVar.Feedback = 0;
	hdacPatchVar.Counter = 0;
	hdacPatchVar.Increaser = 0;

	KMS_DEBUG("hdac cts patch unload!\n");
}

void via_chrome9_hdac_cts_patch_open(void)
{
	KMS_DEBUG("try to open\n");
	if (hdac_timer_open == 0) {
		hdac_timer_open = 1;
		via_chrome9_hdac_cts_patch_timer_init();
	}
}

void via_chrome9_hdac_cts_patch_close(void)
{
	KMS_DEBUG("try to close\n");
	if (hdac_timer_open) {
		hdac_timer_open = 0;
		via_chrome9_hdac_cts_patch_timer_exit();
	}
}

