/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice
 * (including the next paragraph) shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL VIA, S3 GRAPHICS, AND/OR
 * ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
  * Programming info contained in this module for AD9389
  * can be found in
  * http://www.classic.nxp.com/acrobat_download2/literature/9398/39340011.pdf 
  */

#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/io.h>
#include <linux/slab.h>

#include "via_chrome9_drv.h"
#include "via_chrome9_i2c.h"

#define CHAN_MAX	(7)

static struct via_chrome9_i2c_chan *via_chrome9_i2c_chans[CHAN_MAX] = {NULL};

static u8 read_mask_from_i2c_reg(u16 io_name, u8 mask)
{
	u8 val = 0;
	val = via_chrome9_read_vga_io(io_name);
	val = val & mask;
	return val;
}

static void write_mask_to_i2c_reg(u16 io_name, u8 mask, u8 value)
{
	via_chrome9_write_vga_io_bits(io_name, value, mask);
}

static void set_sda(void *priv, int state)
{
	struct via_chrome9_i2c_chan *chan = priv;
	struct via_chrome9_i2c_hw_rec *hw = chan->data;
	if (state)
		write_mask_to_i2c_reg(hw->i2c_reg, hw->set_sda_mask,
			hw->set_sda_mask);
	else
		write_mask_to_i2c_reg(hw->i2c_reg, hw->set_sda_mask, 0);
}

static void set_scl(void *priv, int state)
{
	struct via_chrome9_i2c_chan *chan = priv;
	struct via_chrome9_i2c_hw_rec *hw = chan->data;
	if (state)
		write_mask_to_i2c_reg(hw->i2c_reg, hw->set_scl_mask,
			hw->set_scl_mask);
	else
		write_mask_to_i2c_reg(hw->i2c_reg, hw->set_scl_mask, 0);
}

static int get_scl(void *priv)
{
	struct via_chrome9_i2c_chan *chan = priv;
	struct via_chrome9_i2c_hw_rec *hw = chan->data;

	return read_mask_from_i2c_reg(hw->i2c_reg, hw->get_scl_mask);
}

static int get_sda(void *priv)
{
	struct via_chrome9_i2c_chan *chan = priv;
	struct via_chrome9_i2c_hw_rec *hw = chan->data;

	return read_mask_from_i2c_reg(hw->i2c_reg, hw->get_sda_mask);
}

static void via_set_sda(void *priv, int value, int ack)
{
	struct via_chrome9_i2c_chan *chan = priv;
	struct via_chrome9_i2c_hw_rec *hw = chan->data;

	if (ack) {
		write_mask_to_i2c_reg(hw->i2c_reg, hw->sda_ctl_mask, 0);
	} else {
		write_mask_to_i2c_reg(hw->i2c_reg, hw->sda_ctl_mask,
			hw->sda_ctl_mask);
		set_sda(priv, value);
	}
}

static void via_set_scl(void *priv, int value)
{
	set_scl(priv, value);
}

static inline void via_set_sda_lo(struct i2c_algo_bit_data *adap)
{
	via_set_sda(adap->data, 0, 0);
	udelay((adap->udelay + 1) / 2);
}

static inline void via_set_sda_hi(struct i2c_algo_bit_data *adap)
{
	via_set_sda(adap->data, 1, 0);
	udelay((adap->udelay + 1) / 2);
}

static inline void via_set_scl_lo(struct i2c_algo_bit_data *adap)
{
	via_set_scl(adap->data, 0);
	udelay(adap->udelay / 2);
}

/*
 * Raise scl line, and do checking for delays. This is necessary for slower
 * devices.
 */
static int via_set_scl_hi(struct i2c_algo_bit_data *adap)
{
	unsigned long start;

	via_set_scl(adap->data, 1);

	/* Not all adapters have scl sense line... */
	if (!adap->getscl)
		goto done;

	start = jiffies;
	while (!get_scl(adap->data)) {
		/* This hw knows how to read the clock line, so we wait
		 * until it actually gets high.  This is safer as some
		 * chips may hold it low ("clock stretching") while they
		 * are processing data internally.
		 */
		if (time_after(jiffies, start + adap->timeout))
			return -ETIMEDOUT;
		cond_resched();
	}
#ifdef DEBUG
	if (jiffies != start && i2c_debug >= 3)
		pr_debug("i2c-algo-bit: needed %ld jiffies for SCL to go "
			 "high\n", jiffies - start);
#endif

done:
	udelay(adap->udelay);
	return 0;
}


/* --- other auxiliary functions --------------------------------------	*/
static void via_i2c_start(struct i2c_algo_bit_data *adap)
{
	/* assert: scl, sda are high */
	via_set_sda(adap->data, 0, 0);
	udelay(adap->udelay);
	via_set_scl(adap->data, 0);
	udelay(adap->udelay);
}

static void via_i2c_repstart(struct i2c_algo_bit_data *adap)
{
	/* assert: scl is low */
	via_set_sda(adap->data, 1, 0);
	via_set_scl_hi(adap);
	via_set_sda(adap->data, 0, 0);
	udelay(adap->udelay);
	via_set_scl_lo(adap);

}


static void via_i2c_stop(struct i2c_algo_bit_data *adap)
{
	/* assert: scl is low */
	via_set_sda_lo(adap);
	via_set_scl_hi(adap);
	via_set_sda(adap->data, 1, 0);
	udelay(adap->udelay);
}



/* send a byte without start cond., look for arbitration,
   check ackn. from slave */
/* returns:
 * 1 if the device acknowledged
 * 0 if the device did not ack
 * -ETIMEDOUT if an error occurred (while raising the scl line)
 */
static int via_i2c_outb(struct i2c_adapter *i2c_adap, unsigned char c)
{
	int i;
	int sb;
	int ack;
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;
	/* assert: scl is low */
	for (i = 7; i >= 0; i--) {
		sb = (c >> i) & 1;
		via_set_sda(adap->data, sb, 0);
		udelay((adap->udelay + 1) / 2);
		if (via_set_scl_hi(adap) < 0) { /* timed out */
			KMS_DEBUG("via_i2c_outb: 0x%02x, timeout at bit #%d\n",
					(int)c, i);
			return -ETIMEDOUT;
		}
		/* FIXME do arbitration here:
		 * if (sb && !getsda(adap)) -> ouch! Get out of here.
		 *
		 * Report a unique code, so higher level code can retry
		 * the whole (combined) message and *NOT* issue STOP.
		 */
		via_set_scl_lo(adap);
	}
	via_set_sda(adap->data, 1, 1);

	/* timeout */
	if (via_set_scl_hi(adap) < 0) {
		KMS_DEBUG("via_i2c_outb: 0x%02x, timeout at ack\n",	(int)c);
		return -ETIMEDOUT;
	}

	/* read ack: SDA should be pulled down by slave, or it may
	 * NAK (usually to report problems with the data we wrote).
	 */

	/* ack: sda is pulled low -> success */
	ack = !get_sda(adap->data);
	KMS_DEBUG("via_i2c_outb: 0x%02x %s\n", (int)c, ack ? "A" : "NA");
	via_set_scl_lo(adap);

	/* assert: scl is low (sda undef) */
	return ack;
}


static int via_i2c_inb(struct i2c_adapter *i2c_adap)
{
	/* read byte via i2c port, without start/stop sequence	*/
	/* acknowledge is sent in i2c_read.			*/
	int i;
	unsigned char indata = 0;
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;

	/* assert: scl is low */
	via_set_sda(adap->data, 1, 1);
	udelay(adap->udelay / 2);
	for (i = 0; i < 8; i++) {

		/* timeout */
		if (via_set_scl_hi(adap) < 0) {
			KMS_DEBUG("via_i2c_inb: timeout at bit #%d\n", 7-i);
			return -ETIMEDOUT;
		}

		indata *= 2;
		if (get_sda(adap->data))
			indata |= 0x01;
		via_set_scl(adap->data, 0);
		udelay(i == 7 ? adap->udelay / 2 : adap->udelay);
	}
	/* assert: scl is low */
	return indata;
}

/* ----- Utility functions
 */

/* try_address tries to contact a chip for a number of
 * times before it gives up.
 * return values:
 * 1 chip answered
 * 0 chip did not answer
 * -x transmission error
 */
static int via_try_address(struct i2c_adapter *i2c_adap,
		unsigned char addr, int retries)
{
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;
	int i, ret = 0;

	for (i = 0; i <= retries; i++) {
		ret = via_i2c_outb(i2c_adap, addr);
		if (ret == 1 || i == retries)
			break;
		KMS_DEBUG("emitting stop condition\n");
		via_i2c_stop(adap);
		udelay(adap->udelay);
		yield();
		KMS_DEBUG("emitting start condition\n");
		via_i2c_start(adap);
	}
	if (i && ret) {
		KMS_DEBUG("Used %d tries to %s client at 0x%02x:%s\n", i+1,
			addr & 1 ? "read from" : "write to", addr >> 1,
			ret == 1 ? "success" : "failed, timeout?");
	}
	return ret;
}

static int via_send_bytes(struct i2c_adapter *i2c_adap, struct i2c_msg *msg)
{
	const unsigned char *temp = msg->buf;
	int count = msg->len;
	unsigned short nak_ok = msg->flags & I2C_M_IGNORE_NAK;
	int retval;
	int wrcount = 0;

	while (count > 0) {
		retval = via_i2c_outb(i2c_adap, *temp);
		/* OK/ACK; or ignored NAK */
		if ((retval > 0) || (nak_ok && (retval == 0))) {
			count--;
			temp++;
			wrcount++;

		/* A slave NAKing the master means the slave didn't like
		 * something about the data it saw.  For example, maybe
		 * the SMBus PEC was wrong.
		 */
		} else if (retval == 0) {
			dev_err(&i2c_adap->dev, "sendbytes: NAK bailout.\n");
			return -EIO;

		/* Timeout; or (someday) lost arbitration
		 *
		 * FIXME Lost ARB implies retrying the transaction from
		 * the first message, after the "winning" master issues
		 * its STOP.  As a rule, upper layer code has no reason
		 * to know or care about this ... it is *NOT* an error.
		 */
		} else {
			dev_err(&i2c_adap->dev, "sendbytes: error %d\n",
					retval);
			return retval;
		}
	}
	return wrcount;
}

static int via_ack_nak(struct i2c_adapter *i2c_adap, int is_ack)
{
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;

	/* assert: sda is high */
	/* send ack */
	if (is_ack)
		via_set_sda(adap->data, 0, 0);
	udelay((adap->udelay + 1) / 2);

	if (via_set_scl_hi(adap) < 0) {
		/* timeout */
		dev_err(&i2c_adap->dev, "readbytes: ack/nak timeout\n");
		return -ETIMEDOUT;
	}
	via_set_scl_lo(adap);
	return 0;
}

static int via_read_bytes(struct i2c_adapter *i2c_adap, struct i2c_msg *msg)
{
	int inval;
	int rdcount = 0;	/* counts bytes read */
	unsigned char *temp = msg->buf;
	int count = msg->len;
	const unsigned flags = msg->flags;

	while (count > 0) {
		inval = via_i2c_inb(i2c_adap);
		if (inval >= 0) {
			*temp = inval;
			rdcount++;
		} else {   /* read timed out */
			break;
		}
		temp++;
		count--;

		/* Some SMBus transactions require that we receive the
		   transaction length as the first read byte. */
		if (rdcount == 1 && (flags & I2C_M_RECV_LEN)) {
			if (inval <= 0 || inval > I2C_SMBUS_BLOCK_MAX) {
				if (!(flags & I2C_M_NO_RD_ACK))
					via_ack_nak(i2c_adap, 0);
				dev_err(&i2c_adap->dev, "readbytes: invalid "
					"block length (%d)\n", inval);
				return -EREMOTEIO;
			}
			/* The original count value accounts for the extra
			   bytes, that is, either 1 for a regular transaction,
			   or 2 for a PEC transaction. */
			count += inval;
			msg->len += inval;
		}
		if (!(flags & I2C_M_NO_RD_ACK)) {
			inval = via_ack_nak(i2c_adap, count);
			if (inval < 0)
				return inval;
		}
	}
	return rdcount;
}

/* doAddress initiates the transfer by generating the start condition (in
 * try_address) and transmits the address in the necessary format to handle
 * reads, writes as well as 10bit-addresses.
 * returns:
 *  0 everything went okay, the chip ack'ed, or IGNORE_NAK flag was set
 * -x an error occurred (like: -EREMOTEIO if the device did not answer, or
 *	-ETIMEDOUT, for example if the lines are stuck...)
 */
static int via_bit_doAddress(struct i2c_adapter *i2c_adap, struct i2c_msg *msg)
{
	unsigned short flags = msg->flags;
	unsigned short nak_ok = msg->flags & I2C_M_IGNORE_NAK;
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;

	unsigned char addr;
	int ret, retries;

	retries = nak_ok ? 0 : i2c_adap->retries;

	if (flags & I2C_M_TEN) {
		/* this case is not debug */
		/* please fix me if anything trouble */
		/* a ten bit address */
		addr = 0xf0 | ((msg->addr >> 7) & 0x03);
		/* try extended address code...*/
		ret = via_try_address(i2c_adap, addr, retries);

		if ((ret != 1) && !nak_ok)  {
			dev_err(&i2c_adap->dev,
				"died at extended address code\n");
			return -EREMOTEIO;
		}
		/* the remaining 8 bit address */
		ret = via_i2c_outb(i2c_adap, msg->addr & 0x7f);

		if ((ret != 1) && !nak_ok) {
			/* the chip did not ack / xmission error occurred */
			dev_err(&i2c_adap->dev, "died at 2nd address code\n");
			return -EREMOTEIO;
		}

		if (flags & I2C_M_RD) {
			via_i2c_repstart(adap);
			/* okay, now switch into reading mode */
			addr |= 0x01;
			ret = via_try_address(i2c_adap, addr, retries);

			if ((ret != 1) && !nak_ok) {
				dev_err(&i2c_adap->dev,
					"died at repeated address code\n");
				return -EREMOTEIO;
			}
		}
	} else {
		/* normal 7bit address */
		addr = msg->addr << 1;

		if (flags & I2C_M_RD)
			addr |= 1;
		if (flags & I2C_M_REV_DIR_ADDR)
			addr ^= 1;

		ret = via_try_address(i2c_adap, addr, retries);

		if ((ret != 1) && !nak_ok)
			return -ENXIO;
	}

	return 0;
}


static int via_bit_xfer(struct i2c_adapter *i2c_adap,
		struct i2c_msg msgs[], int num)
{
	struct i2c_msg *pmsg;
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;
	int i, ret;
	unsigned short nak_ok;

	KMS_DEBUG("enter %s\n", __func__);
	via_i2c_start(adap);
	for (i = 0; i < num; i++) {
		pmsg = &msgs[i];
		nak_ok = pmsg->flags & I2C_M_IGNORE_NAK;
		if (!(pmsg->flags & I2C_M_NOSTART)) {
			if (i)
				via_i2c_repstart(adap);

			ret = via_bit_doAddress(i2c_adap, pmsg);
			if ((ret != 0) && !nak_ok) {
				KMS_DEBUG(
					"NAK frome device addr 0x%02x msg #%d\n",
					msgs[i].addr, i);
				goto bailout;
			}
		}
		if (pmsg->flags & I2C_M_RD) {
			/* read bytes into buffer */
			ret = via_read_bytes(i2c_adap, pmsg);
			if (ret >= 1)
				KMS_DEBUG("read %d byte%s\n", ret,
					ret == 1 ? "" : "s");

			if (ret < pmsg->len) {
				if (ret >= 0)
					ret = -EREMOTEIO;
				goto bailout;
			}
		} else {
			/* write bytes from buffer */
			ret = via_send_bytes(i2c_adap, pmsg);
			if (ret >= 1) {
				KMS_DEBUG("wrote %d bytes %s\n", ret,
					ret == 1 ? "" : "s");
			}
			if (ret < pmsg->len) {
				if (ret >= 0)
					ret = -EREMOTEIO;
				goto bailout;
			}
		}
	}
	ret = i;

bailout:
	KMS_DEBUG("emitting stop condition\n");
	via_i2c_stop(adap);

	return ret;
}

static u32 via_bit_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL |
	       I2C_FUNC_SMBUS_READ_BLOCK_DATA |
	       I2C_FUNC_SMBUS_BLOCK_PROC_CALL |
	       I2C_FUNC_10BIT_ADDR | I2C_FUNC_PROTOCOL_MANGLING;
}


static const struct i2c_algorithm via_i2c_bit_algo = {
	.master_xfer	= via_bit_xfer,
	.functionality	= via_bit_func,
};

static int pre_xfer(struct i2c_adapter *i2c_adap)
{


	return 0;
}

static void post_xfer(struct i2c_adapter *i2c_adap)
{

}

static int get_free_chan(void)
{
	int i;

	for (i = 0; i < CHAN_MAX; i++) {
		if (via_chrome9_i2c_chans[i] == NULL)
			break;
	}

	if (CHAN_MAX == i)
		return -1;

	return i;

}

int via_chrome9_i2c_bit_add_bus(struct i2c_adapter *adap)
{
	/* register new adapter to i2c module... */
	adap->algo = &via_i2c_bit_algo;
	adap->retries = 3;

	return i2c_add_adapter(adap);
}

static int via_chrome9_register_i2c_chan(struct via_chrome9_i2c_chan *chan)
{
	int chan_id;

	chan_id = get_free_chan();
	if (chan_id < 0) {
		DRM_ERROR("%s can not alloc free chan id\n", __func__);
		return -ENODEV;
	}

	chan->chan_id = chan_id;
	via_chrome9_i2c_chans[chan_id] = chan;

	return 0;
}

static void via_chrome9_unregister_i2c_chan(struct via_chrome9_i2c_chan *chan)
{
	int chan_id;

	chan_id = chan->chan_id;
	via_chrome9_i2c_chans[chan_id] = NULL;
}

struct via_chrome9_i2c_chan *via_chrome9_lookup_i2c_chan(char name)
{
	int i;
	struct via_chrome9_i2c_chan *chan;

	for (i = 0; i < CHAN_MAX; i++) {
		chan = via_chrome9_i2c_chans[i];
		if (chan->name == name)
			return chan;
	}

	return NULL;
}

struct via_chrome9_i2c_chan *via_chrome9_i2c_create(struct drm_device *dev,
		struct via_chrome9_i2c_hw_rec *data,
		const char *name)
{
	struct via_chrome9_i2c_chan *chan;
	struct drm_via_chrome9_private *via_private = dev->dev_private;
	int ret;

	DRM_DEBUG("enter %s\n", __func__);
	chan = kzalloc(sizeof(struct via_chrome9_i2c_chan), GFP_KERNEL);
	if (chan == NULL) {
		printk(KERN_ERR "%s err no mem\n", __func__);
		return NULL;
	}

	sprintf(chan->adapter.name, "via i2c bit bus %s", name);
	chan->data = data;
	chan->name = data->name;
	chan->dev = dev;
	chan->adapter.owner = THIS_MODULE;
	chan->iomem = via_private->mmio_map;   /*may fixup later*/
	i2c_set_adapdata(&chan->adapter, chan);

	chan->adapter.algo_data = &chan->bit;
	chan->adapter.class = I2C_CLASS_DDC;
	chan->bit.pre_xfer = pre_xfer;
	chan->bit.post_xfer = post_xfer;
	chan->bit.setsda = set_sda;
	chan->bit.setscl = set_scl;
	chan->bit.getsda = get_sda;
	chan->bit.getscl = get_scl;
	chan->bit.udelay = 15;
	chan->bit.timeout = 2;
	chan->bit.data = chan;
	if (CHAN_FROM_GPIO == data->type) {
		KMS_DEBUG("via i2c bit bus %s", name);
		ret = via_chrome9_i2c_bit_add_bus(&chan->adapter);
	} else {
		ret = i2c_bit_add_bus(&chan->adapter);
	}

	if (ret) {
		DRM_ERROR("Failed to register bit i2c %s\n", name);
		goto out_free;
	}

	ret = via_chrome9_register_i2c_chan(chan);

	if (ret < 0) {
		DRM_ERROR("failed to register i2c chan\n");
		goto register_err;
	}
	return chan;

register_err:
	i2c_del_adapter(&chan->adapter);
out_free:
	kfree(chan);
	return NULL;

}

void via_chrome9_free_i2c_chan(struct via_chrome9_i2c_chan *chan)
{
	via_chrome9_unregister_i2c_chan(chan);
	i2c_del_adapter(&chan->adapter);
	kfree(chan);
}

void via_chrome9_free_all_i2c_chan(void)
{
	int i;

	for (i = 0; i < CHAN_MAX; i++) {
		if (via_chrome9_i2c_chans[i] != NULL)
			via_chrome9_free_i2c_chan(via_chrome9_i2c_chans[i]);
	}
}
void via_chrome9_i2c_readbytes(struct via_chrome9_i2c_chan *chan,
		u8 slave_addr, char offset,
		u8 *buffer, unsigned int size)
{
	u8 out_buf[2];
	u8 in_buf[2];
	struct i2c_msg msgs[] = {
		{
			.addr = slave_addr,
			.flags = 0,
			.len = 1,
			.buf = out_buf,
		},
		{
			.addr = slave_addr,
			.flags = I2C_M_RD,
			.len = size,
			.buf = in_buf,
		}
	};

	out_buf[0] = offset;
	out_buf[1] = 0;

	if (i2c_transfer(&chan->adapter, msgs, 2) != 2)
		printk(KERN_WARNING "%s failed\n", __func__);
	else
		*buffer = in_buf[0];
}

void via_chrome9_i2c_writebytes(struct via_chrome9_i2c_chan *chan,
		u8 slave_addr, char offset,
		u8 *data, unsigned int size)
{
	u8 *out_buf;
	struct i2c_msg msg = {0};

	out_buf = kzalloc(size + 1, GFP_KERNEL);
	out_buf[0] = offset;
	memcpy(&out_buf[1], data, size);
	msg.addr = slave_addr;
	msg.flags = 0;
	msg.len = size + 1;
	msg.buf = out_buf;

	if (i2c_transfer(&chan->adapter, &msg, 1) != 1)
		printk(KERN_WARNING "%s failed\n", __func__);

	kfree(out_buf);
}



