/*
 * Universal Interface for Intel High Definition Audio Codec
 *
 * HD audio interface patch for VIA VT9880/VT9881 codec
 *
 * Copyright (c) 2008-2009 Annie Liu <annieliu@viatech.com>
 *			   Takashi Iwai <tiwai@suse.de>
 *
 *  This driver is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This driver is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

/* * * * * * * * * * * * * * Release History * * * * * * * * * * * * * * * * */
/*                                                                           */
/* 2009-03-03  Annie Liu  Create the basic patch to support VT9170 codec     */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <linux/init.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/asoundef.h>
#include "hda_codec.h"
#include "hda_local.h"

#define CVT_NID	0x02	/* audio converter1 */
#define PIN_NID	0x03	/* HDMI output pin1 */


struct via_hdmi_spec {
	struct hda_multi_out multiout;
	struct hda_pcm pcm_rec;
};

static struct hda_verb viahdmi_basic_init[] = {	
	/* enable digital output on pin widget */	
	{ PIN_NID, AC_VERB_SET_PIN_WIDGET_CONTROL, PIN_OUT },	
	{} /* terminator */
};

/*
 * Digital out
 */
static int via_hdmi_playback_pcm_open(struct hda_pcm_stream *hinfo,
				     struct hda_codec *codec,
				     struct snd_pcm_substream *substream)
{
	struct via_hdmi_spec *spec = codec->spec;
	return snd_hda_multi_out_dig_open(codec, &spec->multiout);
}

static int via_hdmi_playback_pcm_close(struct hda_pcm_stream *hinfo,
				      struct hda_codec *codec,
				      struct snd_pcm_substream *substream)
{
	struct via_hdmi_spec *spec = codec->spec;
	return snd_hda_multi_out_dig_close(codec, &spec->multiout);
}

static int via_hdmi_playback_pcm_prepare(struct hda_pcm_stream *hinfo,
					   struct hda_codec *codec,
					   unsigned int stream_tag,
					   unsigned int format,
					   struct snd_pcm_substream *substream)
{
	struct via_hdmi_spec *spec = codec->spec;
	return snd_hda_multi_out_dig_prepare(codec, &spec->multiout, stream_tag,
					     format, substream);
}

static struct hda_pcm_stream via_hdmi_digital_playback = {
	.substreams = 1,
	.channels_min = 2,
	.channels_max = 2,
	.nid = CVT_NID, /* NID to query formats and rates and setup streams */
	.ops = {
		.open    = via_hdmi_playback_pcm_open,
		.close   = via_hdmi_playback_pcm_close,
		.prepare = via_hdmi_playback_pcm_prepare
	},
};

static int via_hdmi_build_controls(struct hda_codec *codec)
{
	struct via_hdmi_spec *spec = codec->spec;
	int err;

	err = snd_hda_create_spdif_out_ctls(codec, spec->multiout.dig_out_nid,
						   spec->multiout.dig_out_nid);
	if (err < 0)
		return err;

	return 0;
}

static int via_hdmi_build_pcms(struct hda_codec *codec)
{
	struct via_hdmi_spec *spec = codec->spec;
	struct hda_pcm *info = &spec->pcm_rec;

	codec->num_pcms = 1;
	codec->pcm_info = info;

	info->name = "VIA GFX HDMI";
	info->pcm_type = HDA_PCM_TYPE_HDMI;
	info->stream[SNDRV_PCM_STREAM_PLAYBACK] = via_hdmi_digital_playback;

	return 0;
}

static int via_hdmi_init(struct hda_codec *codec)
{
	snd_hda_sequence_write(codec, viahdmi_basic_init);
	return 0;
}

static void via_hdmi_free(struct hda_codec *codec)
{
	kfree(codec->spec);
}

static struct hda_codec_ops via_hdmi_patch_ops = {
	.init			= via_hdmi_init,
	.free			= via_hdmi_free,
	.build_pcms		= via_hdmi_build_pcms,
	.build_controls 	= via_hdmi_build_controls,
};

static int patch_via_hdmi(struct hda_codec *codec)
{
	struct via_hdmi_spec *spec;

	spec = kzalloc(sizeof(*spec), GFP_KERNEL);
	if (spec == NULL)
		return -ENOMEM;

	spec->multiout.num_dacs = 0;	  /* no analog */
	spec->multiout.max_channels = 2;
	spec->multiout.dig_out_nid = CVT_NID;/*pure-digital case*/

	codec->spec = spec;
	codec->patch_ops = via_hdmi_patch_ops;

	return 0;
}

/*
 * patch entries
 */
struct hda_codec_preset snd_hda_preset_viahdmi[] = {
	{ .id = 0x11069f80, .name = "VT9170 9f80", .patch = patch_via_hdmi },
	{ .id = 0x11069f81, .name = "VT9170 9f81", .patch = patch_via_hdmi },
	{} /* terminator */
};

MODULE_ALIAS("snd-hda-codec-id:11069f80");
MODULE_ALIAS("snd-hda-codec-id:11069f81");

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("VIA GFX HDMI HD-audio codec");
MODULE_AUTHOR("VIA TECH");
MODULE_VERSION("1.1.1");

static struct hda_codec_preset_list viahdmi_list = {
	.preset = snd_hda_preset_viahdmi,
	.owner = THIS_MODULE,
};

static int __init patch_viahdmi_init(void)
{
        printk(KERN_ALERT "snd_hda_codec_viahdmi driver 1.1.1 loaded.\n");
	return snd_hda_add_codec_preset(&viahdmi_list);
}

static void __exit patch_viahdmi_exit(void)
{
        printk(KERN_ALERT "snd_hda_codec_viahdmi driver 1.1.1 unloaded.\n");
	snd_hda_delete_codec_preset(&viahdmi_list);
}

module_init(patch_viahdmi_init)
module_exit(patch_viahdmi_exit)
